#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <getopt.h>
#include <omp.h>
#include "fire.h"

int forest_size;
int trees;

void usage() {
    fprintf(stderr,"Usage: %s -p <burn-probability> -n <size-forest>, where\n \
\t[-p float] probability that a green tree will catch fire due to burning neighnor, between 0 and 1\n \
\t[-n int] the number of trees per side of a square forest, between 2 and 80\n",
            program_invocation_short_name);
}

/* This function processes the two user-supplied-values handed to the program at runtime.
 * The burn probability (-p) and the forest size (-n) must be supplied, else the program
 * will print out an explanatory note and abort.*/
struct user_args get_user_args(const int argc,char **argv) {
    int opt, tmp;
    struct user_args args;

    args.burn_ratio = 0;
    /* While there are parameter flags (-p and-n in our case) not-yet-processed, keep looping. */
    while((opt = getopt(argc,argv,":p:n:")) != -1) {
        switch(opt) {
            case 'p':
		/* If the flag is -p, grab the value that follows and ensure it's not too big; 
		if it is, complain and abort. Otherwise, use that value as the burn probability */
                if(atof(optarg) == HUGE_VAL) {
                    fprintf(stderr,"Received invalid value: %s, %s:%d\n",
                            strerror(errno),__FILE__,__LINE__);
                    usage();
                    abort();
                }

    		if(atof(optarg) <= 0 || atof(optarg) > 1) {
			fprintf(stderr,"Supply valid burn ratio!\n");
			usage();
			abort();
    		}

                args.burn_ratio = atof(optarg);
               break; 
	    case 'n':
		/* If the flag is -n, grab the value that follows and ensure it's not too big or small;
		if it is, complain and abort. Otherwise, use that value as the number of trees per side 
		of the square forest. For instance, using "-n 20" yields a 20x20 forest of trees. */
		tmp = atof(optarg);
		if(tmp>80 || tmp<2){
                    fprintf(stderr,"Received invalid value: %s, %s:%d\n",
                            strerror(errno),__FILE__,__LINE__);
                    usage();
                    abort();
		}
		forest_size = tmp;
		trees = forest_size*forest_size;
		break;
	    default:
		/* If the flag is neither, something nefarious is afoot. Abort immediately. */
                fprintf(stderr,"Received invalid flag: %s, %s:%d\n",strerror(errno),__FILE__,__LINE__);
		usage();
		abort();
        }
    }


    return args;
}

/* This function uses the malloc function to allocate memory for the forests, which are composed of
   trees. */
struct forest *init_forest(
        const struct user_args args,
        struct drand48_data *rand_buffer,
	int startBurn) {

    struct forest *burning_forest;
    size_t forest_bytes = sizeof(forest) * trees;
    double rand_result;
    unsigned int i;
    int up,down,right,left;
    /* This is how to allocate memory for a variable using the malloc function. It is in the if statement to test 
       if the allocation worked; if not, print an explanation to the user and abort. */ 
    if((burning_forest = malloc(forest_bytes)) == NULL) {
        fprintf(stderr,"Failed to malloc %zd bytes for forest: %s, %s:%d\n", forest_bytes,strerror(errno),__FILE__,__LINE__);
	return NULL;
    }

    /* This loops through each tree of the forest like an excited orangutan. It initializes each tree 
       as being NOT on fire (GREEN_TREE is defined as 0 above), determines the location (by index in the array) 
       of the neighbors to the north, south, east, and west. */
#pragma omp parallel for private(up, down, left, right)
    for(i=0;i<trees;i++){
	burning_forest[i].tree = GREEN_TREE;

	up = i-forest_size;
	down = i+forest_size;
        left = i-1;
        right = i+1;

        burning_forest[i].neighbors.up = up >= 0
            ? up
            : OOB;

        burning_forest[i].neighbors.left = (left >= 0 && i%forest_size != 0)
            ? left
            : OOB;

        burning_forest[i].neighbors.down = down < trees
            ? down
            : OOB;

        burning_forest[i].neighbors.right = (right < trees && i%forest_size != forest_size-1)
            ? right
            : OOB;
    }

    /* If the forest being initialized is the 'next_burning_forest' (it is the one in which we compute which trees 
       are catching fire in the upcoming timestep), generate a random number to determine which tree will be first 
       lit on fire. Increment that tree's "burn count" to start it burning. */
    if(startBurn == 1){
    	drand48_r(rand_buffer,&rand_result);
    	i = (int)(rand_result*trees);
    	burning_forest[i].tree++;
    }

    return burning_forest;
}

/* This function (burn_trees) loops through all the trees and assigns the state of the tree in the "upcoming" timestep 
   (calculated and stored in next_burning_forest) to the trees (in burning_forest). */
int burn_trees(
        const struct user_args args,
        struct forest *burning_forest,
        struct forest *next_burning_forest,
        struct drand48_data *rand_buffer
        ) {
   unsigned int i,j,t,cur_burn_trees,started_burn_trees; 

   /* The simulation is over when no burning trees remain */
   t = 0;
   do {
	cur_burn_trees = started_burn_trees = 0;
	
	/* Loop through all the trees, assign the computed values for the upcoming timestep into the current burning_forest. */
#pragma omp parallel for 
	for(i=0;i<trees;i++) {
		burning_forest[i].tree = next_burning_forest[i].tree;
	}

	/* Loop through all the updated tree values, if one is noticed to be burning, count it and call the burn_neighbors 
	   function (which will give it a chance to spread the flame to neighboring trees not on fire). */
#pragma omp parallel for 
	for(i=0;i<trees;i++) {
		if(burning_forest[i].tree > GREEN_TREE && burning_forest[i].tree < BLACK_TREE) {
                	next_burning_forest[i].tree++;
                	cur_burn_trees++;
                	started_burn_trees += burn_neighbors(args,burning_forest,next_burning_forest,i,rand_buffer);
		}
	}

	if(t==0){fprintf(stderr,"T:0, 1 tree is on fire\n");}

	t++;
	
	/* This nicely prints out the forest to the screen t each timestep */
	/* This printing can be done by threads easily because order doesn't matter */
#pragma omp parallel for 
	for(i=0;i<forest_size+1;i++){fprintf(stderr, "__");}

	fprintf(stderr,"_\n");
	
	for(i=0;i<forest_size;i++){
		fprintf(stderr, "| ");
		for(j=0;j<forest_size;j++){
			fprintf(stderr,"%u ",burning_forest[i*forest_size+j].tree);
		}
		fprintf(stderr,"|\n");
	}
	
	/* This printing can be done by threads easily because order doesn't matter */
#pragma omp parallel for 
	for(i=0;i<2*forest_size+3;i++){fprintf(stderr, "-");}
	fprintf(stderr,"\n");


	fprintf(stderr, "T:%u, %u trees started burning this time step, %u were already burning\n", t, started_burn_trees,cur_burn_trees);

	usleep(100000);/* This slows the display of the forest to make it look like animation */


   } while((cur_burn_trees+started_burn_trees) > 0);
   /* Continue looping to calculate new forests as long as there are trees burning. */
   return 0;
}

/* This function sees if the given tree on fire catches any neighbors on fire. */
unsigned short int burn_neighbors(
        const struct user_args args,
        struct forest *burning_forest,
        struct forest *next_burning_forest,
        const unsigned int i,
        struct drand48_data *rand_buffer
        ) {
    unsigned short int started_burn_trees = 0;
    double rand_result;

    /* This big nested if statement tests if the current tree can set a particular neighbor (the one above) on fire.
       This is repeated for the other 3 neighbors to see if they will be caught on fire too. */
    if(burning_forest[i].neighbors.up != OOB && burning_forest[burning_forest[i].neighbors.up].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
	    if(next_burning_forest[burning_forest[i].neighbors.up].tree==0){
		started_burn_trees++;
	    }
            next_burning_forest[burning_forest[i].neighbors.up].tree=1;
        }
    }

    if(burning_forest[i].neighbors.down != OOB
            && burning_forest[burning_forest[i].neighbors.down].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
	    if(next_burning_forest[burning_forest[i].neighbors.down].tree==0){
		started_burn_trees++;
	    }
            next_burning_forest[burning_forest[i].neighbors.down].tree=1;
        }
    }

    if(burning_forest[i].neighbors.left != OOB
            && burning_forest[burning_forest[i].neighbors.left].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
	    if(next_burning_forest[burning_forest[i].neighbors.left].tree==0){
		started_burn_trees++;
	    }
            next_burning_forest[burning_forest[i].neighbors.left].tree=1;
        }
    }
    if(burning_forest[i].neighbors.right != OOB
            && burning_forest[
                burning_forest[i].neighbors.right].tree == GREEN_TREE) {
        drand48_r(rand_buffer,&rand_result);
        if(rand_result < args.burn_ratio) {
	    if(next_burning_forest[burning_forest[i].neighbors.right].tree==0){
		started_burn_trees++;
	    }
            next_burning_forest[burning_forest[i].neighbors.right].tree=1;
        }
    }

    return started_burn_trees;
}

